import os

class Context:
    """
    A class defining the context for the DOLOST application.

    Attributes:
        base_dir (str): The absolute path of the base directory.
        CONFIG_FOLDER (str): The folder name for Docker client configuration.
        CONFIG_FILE (str): The configuration file name.
        CONFIG_PATH (str): The absolute path to the configuration file.
    """
    # Define the base directory path as a class attribute
    base_dir = os.path.abspath(os.path.dirname(__file__))
    CONFIG_FOLDER = 'docker_client_config'
    CONFIG_FILE = 'config.json'
    CONFIG_PATH = os.path.abspath(os.path.join(CONFIG_FOLDER, CONFIG_FILE))


from .logger import DolostLogger
from .docker_manager import DockerManager
import json
from datetime import datetime

logger = DolostLogger.get_instance()
docker_manager = DockerManager.get_instance()


class ActivityViewer:
    @staticmethod
    def review_logs():
        """
        Review logs from the activity viewer.

        This method retrieves the latest log entries from the activity viewer, specifically focusing on the logs generated by the decoys. It connects to the DolosT-Collector container, which serves as the centralized log collector for all decoys. It then tail the last 50 lines of each log file within the '/var/log/decoys/' folder, continuously monitoring for new log entries.

        Returns:
            list: A list containing the latest log entries from the activity viewer.
        """
        container_id = "DolosT-Collector"
        folder_path = "/var/log/decoys/"

        new_logs = []
        # Create a command to monitor each file within the folder continuously
        command = f"sh -c 'tail -n 50 {folder_path}*'"

        filters = {'name': [container_id]}
        CollectorExist = docker_manager.client.api.containers(filters=filters)

        # If collector doesnt exist, show a message
        if (CollectorExist):
            # Create the exec instance in the container
            exec_id = docker_manager.client.api.exec_create(container_id, command, tty=True)

            # Start streaming the output of the command
            for line in docker_manager.client.api.exec_start(exec_id['Id'], stream=True):
                log_line = line.decode().strip()  # Remove leading/trailing whitespace
                new_logs.append(log_line)

            # If no logs received, show a message
            if "tail: can't open '/var/log/decoys/*': No such file or directory\r\ntail: no files" in new_logs:
                new_logs = ["No decoy logs received in the environment yet"]
        else:
            new_logs = ["No collector deployed in the environment yet"]

        return new_logs
    
    def review_observable_ips():
        """
        Review logs from the activity viewer.

        This method retrieves the latest log entries from the activity viewer, specifically focusing on the logs generated by the decoys. It connects to the DolosT-Collector container, which serves as the centralized log collector for all decoys. It then tail the last 50 lines of each log file within the '/var/log/decoys/' folder, continuously monitoring for new log entries.

        Returns:
            list: A list containing the latest log entries from the activity viewer.
        """
        container_id = "DolosT-Collector"
        file_path = "/var/log/observables/observable_ips.log"
        observable_ips = []
        new_ips = []

        # Create a command to check for ips on each file
        command = f"sh -c 'tail -n 5000 {file_path}*'"
        filters = {'name': [container_id]}
        CollectorExist = docker_manager.client.api.containers(filters=filters)

        # If collector doesnt exist, show a message
        if (CollectorExist):
            # Create the exec instance in the container
            exec_id = docker_manager.client.api.exec_create(container_id, command, tty=True)

            # Start streaming the output of the command
            for line in docker_manager.client.api.exec_start(exec_id['Id'], stream=True):
                decoded_line = line.decode('utf-8')
                if decoded_line == ("tail: can't open '/var/log/observables/observable_ips.log*': No such file or directory\r\ntail: no files\r\n") :
                    new_ip = '{"id": 1, "ip": "-- No Observable IP --","timestamp": "----" }'
                    observable_ips.append(new_ip)
                else:    
                    records = decoded_line.strip().split('\r\n')
                    # Process each record
                    i = 1
                    for record in records:
                        # Split each record by space to separate the timestamp and the IP address
                        timestamp, ip_address = record.rsplit(' ', 1)
                        observable_ips.append('{"id": '+ str(i) +', "ip": "'+ ip_address +'", "timestamp": "' + timestamp + '"}')
                        if i >= 3:
                            break

                        i = i + 1
        else:
            new_ip = '{"id": 1, "ip": "-- No Collector available --","timestamp": "----" }'
            observable_ips.append(new_ip)

        return observable_ips

    def turn_on_crond():
        """
        This is an auxiliary task to run crond in the collector

        Returns:
            none
        """
        container_id = "DolosT-Collector"

        # Create a command to run crond
        command = "sh -c 'crond'"
        filters = {'name': [container_id]}
        CollectorExist = docker_manager.client.api.containers(filters=filters)

        # If collector doesnt exist, show a message
        if (CollectorExist):
            # Create the exec instance in the container
            exec_id = docker_manager.client.api.exec_create(container_id, command, tty=True)
            docker_manager.client.api.exec_start(exec_id['Id'], stream=True)



from flask_socketio import Namespace, emit
from ..services.logger import DolostLogger
from ..services.docker_manager import ContainerStatsManager
from ..services.activity import ActivityViewer
import threading

class WebSocketNamespace(Namespace):
    """Namespace for WebSocket connections."""

    def __init__(self, namespace=None):
        """
        Initialize the WebSocketNamespace.

        :param str namespace: The namespace for the WebSocket.

        """
        super().__init__(namespace)
        self.logger = DolostLogger.get_instance()
        self.activity_logs_thread = None
        self.thread_stop_event = threading.Event()  # Event to signal thread termination

    def on_connect(self, *args):
        """
        Callback function triggered when a client connects.

        :param \*args: Variable length argument list.

        """
        self.logger.trace("[WS] Client connected", ws=False)
        emit('heartbeat', {'data': 'Connected'})

    def on_disconnect(self, *args):
        """
        Callback function triggered when a client disconnects.

        :param \*args: Variable length argument list.

        """
        self.logger.trace("[WS] Client disconnected", ws=False)
        if self.activity_logs_thread:
            self.thread_stop_event.set()

    def on_request_data(self, json_data):
        """
        Callback function triggered on request for data.

        :param dict json_data: JSON object containing the request data.
            Should contain 'container_id' key.

        """
        if 'container_id' in json_data and isinstance(json_data['container_id'], str):
            container_id = json_data['container_id']
            data = ContainerStatsManager().fetch_container_stats(container_id)
            emit('update_data', data)
        else:
            emit('update_data_error', {'error': 'Invalid or missing container_id'})

    def send_activity_logs(self):
        """
        Send activity logs to clients.

        Continuously fetches and emits activity logs until stop event is set.

        """
        auxcount = 0
        while not self.thread_stop_event.is_set():
            logs = ActivityViewer.review_logs()
            self.emit('activity_logs', {'logs': logs})

            observable_ips = ActivityViewer.review_observable_ips()
            self.emit('activity_observable_ips', {'observable_ips': observable_ips})
            
            self.thread_stop_event.wait(5)  # Wait for 5 second before sending next update

            # Auxiliary function turn on crond every 60 seconds
            if (auxcount == 0):
                ActivityViewer.turn_on_crond()
                auxcount = 60
            auxcount = auxcount - 1

    def on_request_activity(self):
        """
        Callback function triggered on request for activity logs.

        Starts a new thread to continuously send activity logs to clients.

        """
        if not self.activity_logs_thread or not self.activity_logs_thread.is_alive():
            # Start a new thread only if it's not already running
            self.thread_stop_event.clear()  # Clear the event flag
            self.activity_logs_thread = threading.Thread(target=self.send_activity_logs)
            self.activity_logs_thread.start()



import argparse
from .app import main as start

def main():
    """
    Function to parse command-line arguments and start the DOLOST framework.

    Command-line arguments:
        --host, -H: Host to run the server on (default: '127.0.0.1')
        --port, -p: Port to run the server on (default: 9874)
        --verbosity, -v: Verbosity level ('TRACE', 'DEBUG', 'INFO', 'WARN', 'ERROR') (default: 'INFO')
        --docker-client, -dc: Docker client type ('tcp', 'tcp_ssl', 'socket', 'env') (default: 'env')

        TCP options:
            --tcp-host: Host for TCP connection (required for --docker-client tcp)
            --tcp-port: Port for TCP connection (required for --docker-client tcp)

        TCP+SSL options:
            --ssl-host: Host for TCP+SSL connection (required for --docker-client tcp_ssl)
            --ssl-port: Port for TCP+SSL connection (required for --docker-client tcp_ssl)
            --ssl-cert: Path to client certificate (required for --docker-client tcp_ssl)
            --ssl-key: Path to client key (required for --docker-client tcp_ssl)
            --ssl-ca: Path to CA certificate (required for --docker-client tcp_ssl)

        Socket options:
            --socket-path: Path to Docker socket (required for --docker-client socket)
    """
    parser = argparse.ArgumentParser(description='DOLOST CLI')
    parser.add_argument('--host', '-H', default='127.0.0.1', help='Host to run the server on')
    parser.add_argument('--port', '-p', type=int, default=9874, help='Port to run the server on')
    parser.add_argument('--verbosity', '-v', choices=['TRACE', 'DEBUG', 'INFO', 'WARN', 'ERROR'], default='INFO', help='Verbosity level')
    parser.add_argument('--docker-client', '-dc', default='env', choices=['tcp', 'tcp_ssl', 'socket', 'env'], help='Docker client type')

    # TCP options
    parser.add_argument('--tcp-host', default='127.0.0.1', help='Host for TCP connection (required for --docker-client tcp)')
    parser.add_argument('--tcp-port', type=int, default=2376, help='Port for TCP connection (required for --docker-client tcp)')

    # TCP+SSL options
    parser.add_argument('--ssl-host', default='127.0.0.1', help='Host for TCP+SSL connection (required for --docker-client tcp_ssl)')
    parser.add_argument('--ssl-port', type=int, default=2376, help='Port for TCP+SSL connection (required for --docker-client tcp_ssl)')
    parser.add_argument('--ssl-cert', default='cert.pem', help='Path to client certificate (required for --docker-client tcp_ssl)')
    parser.add_argument('--ssl-key', default='key.pem', help='Path to client key (required for --docker-client tcp_ssl)')
    parser.add_argument('--ssl-ca', default='ca.pem', help='Path to CA certificate (required for --docker-client tcp_ssl)')

    # Socket options
    parser.add_argument('--socket-path', default='unix:///var/run/docker.sock', help='Path to Docker socket (required for --docker-client socket)')

    args = parser.parse_args()

    docker_client_config = {}

    if args.docker_client == 'tcp':
        docker_client_config = {'tcp': f"tcp://{args.tcp_host}:{args.tcp_port}"}
    
    elif args.docker_client == 'tcp_ssl':
        docker_client_config['tcp_ssl'] = {'host': args.ssl_host, 'port': args.ssl_port, 'cert_path': args.ssl_cert, 'key_path': args.ssl_key, 'ca_path': args.ssl_ca}
    
    elif args.docker_client == 'socket':
        docker_client_config = {'socket': args.socket_path}
    
    elif args.docker_client == 'env':
        docker_client_config = {'from_env': True}

    start(host=args.host, port=args.port, verbosity=args.verbosity, docker_client=docker_client_config)

if __name__ == '__main__':
    main()


import docker
import os, shutil, tempfile, ast, json
from .logger import DolostLogger
from ..context import Context

class DockerManager:
    """
    A class for managing the interactions with docker's API

    Attributes:
        _instance (cls): Instance of the class.
    """
    _instance = None

    @classmethod
    def get_instance(cls, docker_client=None):
        """
        Get an instance of the DockerManager class.

        If an instance does not exist, create a new one. If a `docker_client` is provided,
        configure the Docker client and check the connection.

        Args:
            docker_client (docker.client.DockerClient, optional): An instance of DockerClient to use for interaction.

        Returns:
            DockerManager: An instance of the DockerManager class.
        """
        if cls._instance is None:
            cls._instance = cls(docker_client=docker_client)
        
        elif docker_client is not None:
            cls._instance.configure_client(docker_client)
            cls._instance.check_connection()
        return cls._instance

    def __init__(self, docker_client=None):
        """
        Initialize a DockerManager instance.

        Args:
            docker_client: The way to connect to docker
        Returns:
            None
        """
        self.logger = DolostLogger.get_instance()
        self.client = None
        if docker_client:
            self.configure_client(docker_client)

    def check_connection(self):
        """
        Check if the Docker connection is successful.
        """
        if self.client is None:
            self.logger.error("[Docker] client is not set")

        try:
            self.client.ping()
            self.logger.debug("[Docker] connection successful")
            return True
        except Exception as e:
            self.logger.error(f"Failed to connect to Docker: {e}")
            return False

    def configure_client(self, docker_client=None):
        """
        Configure Docker client based on the provided configuration.

        Args:
            docker_client: Configuration for connecting to Docker client
        Returns:
            None
        """
        try:

            if 'from_env' in docker_client:
                self.client = docker.from_env()
                self.logger.trace("[Docker] client configured from ENV")

            elif 'tcp' in docker_client:
                self.client = docker.DockerClient(base_url=docker_client['tcp'])
                self.logger.trace("[Docker] client configured from TCP")

            elif 'ssl' in docker_client:
                tls_config = docker.tls.TLSConfig(
                    client_cert=(docker_client['tcp_ssl']['cert_path'], docker_client['tcp_ssl']['key_path']),
                    ca_cert=docker_client['tcp_ssl']['ca_path']
                )
                self.client = docker.DockerClient(base_url=f"tcp://{docker_client['tcp_ssl']['host']}:{docker_client['tcp_ssl']['port']}", tls=tls_config)
                self.logger.trace("[Docker] client configured from TCP+SSL")

            elif 'socket' in docker_client:
                self.client = docker.DockerClient(base_url=docker_client['socket'])
                self.logger.trace("[Docker] client configured from SOCKET")

            elif docker_client != None:
               # Check if the config folder and file exist
                self.logger.trace(f"[Docker] config file: {Context.CONFIG_PATH}")
                if os.path.exists(Context.CONFIG_PATH):
                    with open(Context.CONFIG_PATH, 'r') as config_file:
                        docker_client = json.load(config_file)
                        self.configure_client(docker_client=docker_client)
                        self.logger.trace("[Docker] client configuration loaded from config file")
                else:
                    self.logger.info("Config file not found. Using default configuration.")
                    docker_client = None
            else:
                raise ValueError("Invalid Docker client configuration")

            # Create the config folder if it doesn't exist
            if not os.path.exists(Context.CONFIG_FOLDER):
                os.makedirs(Context.CONFIG_FOLDER)

            # Write the docker_client data to config.json
            with open(Context.CONFIG_PATH, 'w') as config_file:
                json.dump(docker_client, config_file, indent=4)

        except Exception as e:
            self.logger.error(f"Failed to configure Docker client: {e}")
            self.client = None

    def check_client_configuration(self, docker_client):
        """
        Check before configuring Docker client based on the provided configuration.

        Args:
            docker_client: Configuration for connecting to Docker client
        Returns:
            None
        """
        try:
            if 'from_env' in docker_client:
                client = docker.from_env()
                self.logger.trace("Docker client configured from env")

            elif 'tcp' in docker_client:
                client = docker.DockerClient(base_url=docker_client['tcp'])
                self.logger.trace("Docker client configured from TCP")

            elif 'ssl' in docker_client:
                tls_config = docker.tls.TLSConfig(
                    client_cert=(docker_client['tcp_ssl']['cert_path'], docker_client['tcp_ssl']['key_path']),
                    ca_cert=docker_client['tcp_ssl']['ca_path']
                )
                client = docker.DockerClient(base_url=f"tcp://{docker_client['tcp_ssl']['host']}:{docker_client['tcp_ssl']['port']}", tls=tls_config)
                self.logger.trace("Docker client configured from TCP+SSL")

            elif 'socket' in docker_client:
                client = docker.DockerClient(base_url=docker_client['socket'])
                self.logger.trace("Docker client configured from SOCKET")

            else:
                raise ValueError("Invalid Docker client configuration")
                return False

            client.ping()
            self.logger.debug("Docker connection successful")
            return True
        except Exception as e:
            self.logger.error(f"Failed to configure Docker client: {e}")
            return False

    def get_current_client_config(self):
        """
        Get the current Docker client configuration.

        Returns:
            dict or None: The current Docker client configuration loaded from the config file,
                          or None if the config file does not exist.
        """
        docker_client = None
        if os.path.exists(Context.CONFIG_PATH):
            with open(Context.CONFIG_PATH, 'r') as config_file:
                docker_client = json.load(config_file)
        return docker_client

    def build_context(self, source_service_folder, tmp_location, dockerfile_used):
        """
        Copy files to a temporary folder for Docker build context.

        Args:
            source_service_folder (str): Path to the source directory.
            tmp_location (str): Temporary directory to store the Docker build context.
            dockerfile_used (str): Path to the Dockerfile to be used.

        Returns:
            None
        """
        # Copy files to tmp folder
        tmp_app_destination = os.path.join(tmp_location, "app")

        # Verify if tmp location already exist
        if not os.path.exists(tmp_app_destination):
            os.makedirs(tmp_app_destination)

        # Copy service content to tmp_folder
        for item in os.listdir(source_service_folder):
            s = os.path.join(source_service_folder, item)
            d = os.path.join(tmp_app_destination, item)
            if os.path.isdir(s):
                shutil.copytree(s, d, dirs_exist_ok=True)
            else:
                shutil.copy2(s, d)

        # Copy Dockerfile to tmp_folder
        tmp_dockerfile_destination = tmp_location + "/Dockerfile"
        shutil.copy(dockerfile_used, tmp_dockerfile_destination)

    def build_image(self, image_name=None, decoy_files=None, dockerfile_path=".",):
        """
        Build a Docker image based on the specified Dockerfile and service files.

        Args:
            image_name (str, optional): Name to tag the built image. If None, a default name will be used.
            decoy_files (str, optional): Path to the service files.
            dockerfile_path (str, optional): Path to the Dockerfile. Defaults to '.'.

        Returns:
            str: The image's id.
        """
        source_service_folder = os.path.join(Context.base_dir, f'decoyfiles/{decoy_files}')

        # Temp directory to copy to the Docker image
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            self.build_context(source_service_folder=source_service_folder, tmp_location=tmp_dir_name, dockerfile_used=dockerfile_path)

            if (image_name == "dolost-collector"):
                image_name = "collector_image:dolost"
            else:
                image_name = f"decoy_image_{decoy_files}:dolost"

            try:
                self.logger.info(f"Building Docker image '{image_name}'...")
                self.logger.trace(tmp_dir_name)


                build_output = self.client.api.build(
                    path=tmp_dir_name,
                    tag=image_name,
                    forcerm=True,
                    decode=True
                )
                # Iterate through the output and print progress
                for event in build_output:
                    if 'stream' in event:
                        lines = event['stream'].split('\n')
                        for line in lines:
                            if line.strip():
                                if "Step" in line.strip():
                                    self.logger.debug(line.strip())
                                else:
                                    self.logger.trace(line.strip())

                    if 'aux' in event and 'ID' in event['aux']:
                        image_id = event['aux']['ID']

                self.logger.info(f"Image '{image_name}' successfully built!")
                return image_name

            except docker.errors.BuildError as e:
                self.logger.error(f"Error building Docker image: {e}")

            except docker.errors.APIError as e:
                self.logger.error(f"Error accessing Docker API: {e}")

    def run_container(self, image_name, hostname, name, network_name=None, ipv4_address=None, subnet=None, gateway=None, ports=None ):
        """
        Run a Docker container based on the built image, with custom hostname and container name.

        Args:
            image_name (str): Name of the Docker image to run the container from.
            hostname (str): The container's decoy hostname.
            name (str): The container's name, used for management purposes.
            network_name (str, optional): The name of the network to connect the container to.
            ipv4_address (str, optional): The IPv4 address to assign to the container.
            subnet (str, optional): The subnet address to assign to the container's network.
            gateway (str, optional): The gateway IP address for the container's network.
            ports (dict): The ports configuration for mapping to containers.

        Returns:
            None
        """
        # Rename the name paramenter for the container
        name = f"DolosT-{name}"
        network_name =f"DolosT-{network_name}"
        
        # Check if the container already exist
        dolost_containers = self.get_containers()
        for dolost_container in dolost_containers:
             if name == dolost_container['Names'][0].lstrip('/'):
                self.logger.info(f"Container {name} already deployed. Skiping...")
                return
            
        if network_name:
            filters = {'name': [network_name]}
            network = self.client.api.networks(filters=filters)
            if network:
                self.logger.info(f"Network {network_name} already deployed. Skiping...")
            else:
                self.logger.info(f"Deploying network {network_name} ...")
                # Create the specified internal network if not already exists
                # Define the IPAM config, including the subnet
                ipam_pool = docker.types.IPAMPool(
                    subnet = subnet,
                    gateway = gateway
                )
                ipam_config = docker.types.IPAMConfig(
                    pool_configs=[ipam_pool]
                )
                # Not working BINDING + Internal or Host
                # If the network is not to the Collector leave it with internal access only
                #if (network_name == "DolosTCollectorNetwork"):
                #    NetworkDriver="bridge"
                #else:
                #    NetworkDriver="host"

                network = self.client.api.create_network(network_name, ipam=ipam_config)
                network_id = network["Id"]

        # Define extra_hosts for setting the gateway
        extra_hosts = {hostname: ipv4_address, "gateway": gateway} if network_name else None

        #Convert ports var from str to dict
        ports = ast.literal_eval(ports)  

        log_config = {
            "type": "syslog",
            "config": {
                "syslog-address": f"udp://200.100.0.247:514",
                # "syslog-format": "rfc5424", # Avoid sending the host's hostname
                "tag": hostname,
            }
        }

        # Defining log configuratión and port binding 
        host_config = self.client.api.create_host_config(
            log_config=log_config, 
            port_bindings=ports,
        )

        endpoint_config = self.client.api.create_endpoint_config(
            ipv4_address=ipv4_address,
            
        )
        
        networking_config = self.client.api.create_networking_config({
            network_name: endpoint_config,
        })
        # Continue with deploying the service without network
        container = self.client.api.create_container(
            image_name,
            detach=True,
            hostname=hostname,
            name=name,
            host_config=host_config,
            networking_config=networking_config
            # extra_hosts=extra_hosts,  # Pass extra_hosts here
        )

        #self.client.api.connect_container_to_network(
        #    container=container["Id"], 
        #    net_id=network_id,
        #    ipv4_address=ipv4_address,
        #)

        #Once we have the container, start it
        starting = self.client.api.start(container=container.get('Id'))

        self.logger.debug(f"Container ID: {container.get('Id')}")

    def start(self, container_id):
        """
        Starts a Docker container.

        Args:
            container_id (str): The ID of the container to start.

        Returns:
            None
        """
        try:
            self.client.api.start(container_id)
            self.logger.info(f"Container {container_id} started successfully.")
        except docker.errors.APIError as e:
            self.logger.error(f"Failed to start container {container_id}: {e}")

    def stop(self, container_id):
        """
        Stops a Docker container.

        Args:
            container_id (str): The ID of the container to stop.

        Returns:
            None
        """
        try:
            self.client.api.stop(container_id)
            self.logger.info(f"Container {container_id} stopped successfully.")
        except docker.errors.APIError as e:
            self.logger.error(f"Failed to stop container {container_id}: {e}")
    
    def clean_container(self, name):
        """
        Stop and remove a Docker container with the specified name.

        Args:
            name (str): The name of the Docker container to be cleaned.

        Returns:
            None
        """
        name = f"DolosT-{name}"    
        dolost_containers = self.get_containers()
        for container in dolost_containers:
            if name == container['Names'][0].lstrip('/'):
                self.logger.info(f"Cleaning Docker container '{name}'...")
                self.client.api.remove_container(container=name, force=True)
                self.logger.info(f"Container '{name}' successfully cleaned!")

    def clean_networks(self):
        """
        Remove unused Docker networks with the prefix DolosT-.

        Returns:
            None
        """
        network_name_prefix = "DolosT-"    
        dolost_networks = self.client.networks.list(names="DolosT-*")
        for network in dolost_networks:
            if network.containers == []:
                try:
                    self.client.api.remove_network(network.id)
                    self.logger.info("Network " + network.name + " successfully cleaned!")
                except:
                    self.logger.info("Network " + network.name + " cannot be cleaned, maybe its in use")

    def create_network(self, network_name, subnet, gateway):
        """
        Create a Docker network with the specified name, subnet, and gateway if it does not already exist.

        Args:
            network_name (str): The name of the network.
            subnet (str): The subnet in CIDR notation (e.g., '10.0.0.0/24').
            gateway (str): The gateway IP address for the network.

        Returns:
            None
        """
        try:
            networks = self.client.networks.get(network_name)
        except docker.errors.NotFound:
            ipam_pool = docker.types.IPAMPool(subnet=subnet, gateway=gateway)
            ipam_config = docker.types.IPAMConfig(pool_configs=[ipam_pool])
            self.client.networks.create(network_name, driver="bridge", ipam=ipam_config)

    def connect_to_network(self, container_name, network_name, ipv4_address, gateway):
        """
        Connect a Docker container to a network with the specified name, IP address, and gateway.

        Args:
            container_name (str): The name of the container to connect to the network.
            network_name (str): The name of the network.
            ipv4_address (str): The IPv4 address to assign to the container.
            gateway (str): The gateway IP address for the container's network.

        Returns:
            None
        """
        networks = self.client.networks.get(network_name)

        # Use extra_hosts to set the gateway
        extra_hosts = {container_name: ipv4_address, "gateway": gateway}

        networks.connect(container=container_name, ipv4_address=ipv4_address, extra_hosts=extra_hosts)

    def get_containers(self, all:bool=True):
        """
        Get the docker containers

        Args:
            all (bool): Retrieve all the containers, active or inactive ones.

        Returns:
            containers (list): The list of containers
        """
        return self.client.api.containers(all=all)

class ContainerStatsManager(DockerManager):
    """
    A class for managing container statistics retrieval and calculations.

    Inherits from DockerManager.

    Attributes:
        logger (DolostLogger, optional): The logger instance for logging messages.
        image_name (str, optional): The name of the Docker image.
        decoy_files (str, optional): The name of the service files directory.
        dockerfile_path (str, optional): The path to the Dockerfile.

    Methods:
        fetch_container_stats(container_id):
            Fetches stats for a specific container identified by its ID.

        calculate_cpu_percentage(precpu_stats, cpu_stats):
            Calculates the CPU usage percentage based on the provided CPU stats.

        calculate_memory_percentage(memory_stats):
            Calculates the memory usage percentage based on the provided memory stats.

        parse_container_stats(stats):
            Parses the raw container stats and returns a dictionary containing calculated statistics.
    """
    def __init__(self):
        """
        Initialize a ContainerStatsManager instance.

        Returns:
            None
        """
        docker_manager_instance = DockerManager.get_instance()
        self.logger = docker_manager_instance.logger
        self.client = docker_manager_instance.client

    def fetch_container_stats(self, container_id):
        """
        Fetches stats for a specific container identified by its ID.

        Args:
            container_id (str): The ID of the container.

        Returns:
            dict: A dictionary containing various container statistics.
        """
        stats = self.client.api.stats(container_id, stream=False)
        return self.parse_container_stats(stats)

    def calculate_cpu_percentage(self, precpu_stats, cpu_stats):
        """
        Calculates the CPU usage percentage based on the provided CPU stats.

        Args:
            precpu_stats (dict): The previous CPU stats.
            cpu_stats (dict): The current CPU stats.

        Returns:
            float: The CPU usage percentage.
        """
        cpu_delta = cpu_stats['cpu_usage']['total_usage'] - precpu_stats['cpu_usage']['total_usage']
        system_delta = cpu_stats['system_cpu_usage'] - precpu_stats['system_cpu_usage']
        number_cpus = cpu_stats['online_cpus']

        cpu_percentage = (cpu_delta / system_delta) * number_cpus * 100.0 if system_delta > 0 and cpu_delta > 0 else 0
        return cpu_percentage

    def calculate_memory_percentage(self, memory_stats):
        """
        Calculates the memory usage percentage based on the provided memory stats.

        Args:
            memory_stats (dict): The memory stats.

        Returns:
            float: The memory usage percentage.
        """
        memory_usage = memory_stats['usage']
        memory_limit = memory_stats['limit']

        memory_percentage = (memory_usage / memory_limit) * 100.0 if memory_limit > 0 else 0
        return memory_percentage

    def parse_container_stats(self, stats):
        """
        Parses the raw container stats and returns a dictionary containing calculated statistics.

        Args:
            stats (dict): The raw container stats.

        Returns:
            dict: A dictionary containing various container statistics.
        """
        cpu_percentage = self.calculate_cpu_percentage(stats['precpu_stats'], stats['cpu_stats'])
        memory_percentage = self.calculate_memory_percentage(stats['memory_stats'])

        network_rx = stats['networks']['eth0']['rx_bytes'] if 'networks' in stats else 0
        network_tx = stats['networks']['eth0']['tx_bytes'] if 'networks' in stats else 0

        disk_read, disk_write = 0, 0
        try:
            if 'blkio_stats' in stats and 'io_service_bytes_recursive' in stats['blkio_stats']:
                for stat in stats['blkio_stats']['io_service_bytes_recursive']:
                    if stat['op'] == 'Read':
                        disk_read += stat['value']
                    elif stat['op'] == 'Write':
                        disk_write += stat['value']
        except TypeError:
            pass

        return {
            'cpu_percentage': cpu_percentage,
            'memory_percentage': memory_percentage,
            'network_rx': network_rx,
            'network_tx': network_tx,
            # 'disk_read': disk_read,
            # 'disk_write': disk_write,
        }



from flask import Blueprint, jsonify, render_template, request
from .api import load_operations_db, find_operation_by_id
from ..services.docker_manager import DockerManager
from ..services.logger import DolostLogger
import json
import os


views_blueprint = Blueprint('views', __name__)

logger = DolostLogger.get_instance()
docker_manager = DockerManager.get_instance()

## Frontend Interactions

# Activity view
@views_blueprint.route('/')
@views_blueprint.route('/activity')
def activity():
    """
    Render the activity view.

    Returns:
        str: Rendered HTML template for the activity view.
    """
    return render_template('view_activity.html')

# Operations dashboard
@views_blueprint.route('/operations')
def operations():
    """
    Render the operations dashboard.

    Returns:
        str: Rendered HTML template containing the operations dashboard.
    """
    operations_data = load_operations_db()
    return render_template('operations.html', data=operations_data)

# Endpoint to view operation by ID
@views_blueprint.route('/operations/view/<int:operation_id>', methods=['GET'])
def view_operation(operation_id):
    """
    View details of a specific operation by ID.

    Args:
        operation_id (int): The ID of the operation to view.

    Returns:
        str: Rendered HTML template containing the details of the specified operation.
    """
    try:
        operation = find_operation_by_id(operation_id)

        if operation:
            operation_decoys = operation["decoys"]
            all_containers = docker_manager.get_containers(all=True)
            decoy_names = [entry['Hostname'] for entry in operation_decoys]
            logger.debug(f"Operation's decoy names:  {decoy_names}", ws=False)

            # Create a list to hold all decoys
            operation_decoys_processed = []

            # Process each decoy, check if it's deployed, and format the data
            for decoy in operation_decoys:
                deployed = False
                decoy_data = {
                    'Hostname': f"DolosT-{decoy['Hostname']}",
                    'Description': decoy['Description'],
                    'IP': decoy['IP'],
                    'Subnet': decoy['Subnet'],
                    'Gateway': decoy['Gateway'],
                    'DeceptionNetwork': decoy['DeceptionNetwork'],
                    'ServicePorts': decoy['ServicePorts'],
                    'Service': decoy['Service'],
                    'DecoyFiles': decoy['DecoyFiles'],
                }

                # Check if the decoy has a corresponding container
                for container in all_containers:
                    if f"/DolosT-{decoy['Hostname']}" in container.get('Names', []):
                        deployed = True
                        # Modify the showed stats, from the .json's data to active docker data
                        decoy_data['Id'] = container['Id']
                        decoy_data['Status'] = container['Status']
                        docker_network = "DolosT-" + decoy['DeceptionNetwork']
                        ip_address = container['NetworkSettings']['Networks'][docker_network]['IPAddress']
                        subnet = container['NetworkSettings']['Networks'][docker_network]['IPAddress']
                        netmask = container['NetworkSettings']['Networks'][docker_network]['IPPrefixLen']
                        gateway = container['NetworkSettings']['Networks'][docker_network]['Gateway']
                        
                        decoy_data['IP'] = ip_address if ip_address != "" else "N/A"
                        decoy_data['Subnet'] = f'{subnet}/{netmask}' if subnet != "" else "N/A"
                        decoy_data['Gateway'] = gateway if gateway != "" else "N/A"
                        break

                # If not deployed, set default status and network settings
                if not deployed:
                    decoy_data['Status'] = 'Not Deployed'

                # Add processed decoy data to the list
                operation_decoys_processed.append(decoy_data)

            # Pass the processed decoys to the template
            return render_template('view_operation.html', operation=operation, containers=operation_decoys_processed)
        else:
            return jsonify({'error': 'Operation not found'}), 404
    except Exception as e:
        return jsonify({'error': str(e)}), 500

# Docker client config view
@views_blueprint.route('/config/docker_client')
def docker_client_config():
    """
    Render the Docker client configuration view.

    Returns:
        str: Rendered HTML template for the Docker client configuration view.
    """
    return render_template('docker_client_config.html', docker_config=docker_manager.get_current_client_config())


from flask import Flask
from flask_socketio import SocketIO
from .services.docker_manager import DockerManager
from .services.logger import DolostLogger
from .blueprints.api import api_blueprint
from .blueprints.views import views_blueprint
from .blueprints.websocket import WebSocketNamespace
from .context import Context

def main(verbosity="INFO", docker_client=None, host="127.0.0.1", port:int=9874):
    """
    Main function to start the DOLOST application.

    Args:
        verbosity (str, optional): Verbosity level for logging (default is "INFO").
        docker_client (dict, optional): Configuration for connecting to the Docker client.
        host (str, optional): The host IP address to run the application (default is "127.0.0.1").
        port (int, optional): The port number to run the application (default is 9874).

    Returns:
        None
    """
    app = Flask(__name__)

    production = False
    if production:
        async_mode = 'gevent'
        debug = False
    else:
        async_mode = 'threading'
        debug = True

    socketio = SocketIO(app, async_mode=async_mode)

    # Register Blueprints
    app.register_blueprint(api_blueprint, url_prefix='/api/')
    app.register_blueprint(views_blueprint)

    # Register the WebSocket namespace
    socketio.on_namespace(WebSocketNamespace('/'))

    # Initialize singletons or services
    logger_service = DolostLogger.get_instance(socketio=socketio, level=verbosity)
    docker_manager_service = DockerManager.get_instance(docker_client=docker_client)

    socketio.run(app, debug=debug, host=host, port=port)


import logging
from datetime import datetime
from flask_socketio import SocketIO


class DolostLogger:
	"""
	DolostLogger class provides a simple interface for logging messages with custom log levels
	and emitting those messages to a Socket.IO instance.

	Attributes:
		TRACE (int): Custom log level representing trace messages.
		DEBUG (int): Standard logging.DEBUG level.
		INFO (int): Standard logging.INFO level.
		WARN (int): Custom log level representing warning messages.
		ERROR (int): Standard logging.ERROR level.
		_instance (cls): Instance of the class.

	Methods:
		__init__(self, socketio, level=logging.INFO):
			Initializes a DolostLogger instance.

		_get_numeric_level(self, level):
			Convert a log level from its string representation to the corresponding numeric value.
		
		_register_custom_levels(self):
			Register custom log levels (TRACE and WARN) with the logging module.
		
		_format_log(self, level, message):
			Colorize the log level and message using ANSI escape codes.

		emit_socketio_message(self, message, verbose='INFO'):
			Emits a message to a Socket.IO instance with a specified verbosity level.

		log(self, message, level=logging.INFO):
			Logs a message with a specified log level and emits it to a Socket.IO instance.

		trace(self, message, ws=True):
			Logs a message with the TRACE log level and emits it to a Socket.IO instance.

		debug(self, message, ws=True):
			Logs a message with the DEBUG log level and emits it to a Socket.IO instance.

		info(self, message, ws=True):
			Logs a message with the INFO log level and emits it to a Socket.IO instance.

		warn(self, message, ws=True):
			Logs a message with the WARN log level and emits it to a Socket.IO instance.

		error(self, message, ws=True):
			Logs a message with the ERROR log level and emits it to a Socket.IO instance.
	"""
	TRACE = 5
	DEBUG = logging.DEBUG
	INFO = logging.INFO
	WARN = 35
	ERROR = logging.ERROR
	_instance = None

	@classmethod
	def get_instance(cls, socketio=None, level="TRACE"):
		"""
		Returns the instance of DolostLogger, creating it if necessary.

		Args:
			socketio (Socket.IO): An instance of the Socket.IO class.
			level (str): The log level. Defaults to 'TRACE'.

		Returns:
			DolostLogger: An instance of DolostLogger.

		"""
		if cls._instance is None:
			cls._instance = cls(socketio=socketio, level=level)

		elif socketio is not None:
			cls._instance.socketio = socketio
			cls._instance.logger.setLevel(level)

		return cls._instance

	def __init__(self, socketio: SocketIO, level=logging.INFO):
		"""
		Initializes a DolostLogger instance.

		Args:
			socketio: A Socket.IO instance for emitting messages.
			level (int): The default log level for console prints. Defaults to logging.INFO.
		"""
		self._register_custom_levels()
		self.socketio = socketio
		self.logger = logging.getLogger(__name__)
		self.logger.setLevel(logging.TRACE)

		# Set up a console logger with the specified default level
		console_handler = logging.StreamHandler()
		console_handler.setLevel(self._get_numeric_level(level))

		# formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')

		# console_handler.setFormatter(formatter)
		self.logger.addHandler(console_handler)

	def _get_numeric_level(self, level):
		"""
		Convert a log level from its string representation to the corresponding numeric value.

		Args:
			level (str or int): The log level, either as a string or numeric value.

		Returns:
			int: The numeric representation of the log level.
		"""
		if isinstance(level, str):
			level = getattr(self, level.upper(), logging.INFO)
		return level

	def _register_custom_levels(self):
		"""
		Register custom log levels (TRACE and WARN) with the logging module.

		Returns:
			None
		"""
		# Register custom log levels
		logging.addLevelName(self.TRACE, 'TRACE')
		setattr(logging, 'TRACE', self.TRACE)

		logging.addLevelName(self.WARN, 'WARN')
		setattr(logging, 'WARN', self.WARN)

		logging.getLogger().setLevel(self.TRACE)
		logging.getLogger().setLevel(self.WARN)

	def _format_log(self, level, message):
		"""
		Colorize the log level and message using ANSI escape codes.

		Args:
			level (str): The log level.
			message (str): The log message.

		Returns:
			str: The colorized log level and message.
		"""
		current_time = datetime.now().strftime('%b %d %H:%M:%S')

		# Map log levels to ANSI color codes
		color_mapping = {
			'TRACE': '\033[1;95m',  # COLOR_MAGENTA
			'DEBUG': '\033[1;92m',  # COLOR_CYAN
			'INFO':  '\033[1;96m',  # COLOR_LIGHT_GREEN
			'WARN':  '\033[1;93m',  # COLOR_YELLOW
			'ERROR': '\033[1;91m',  # COLOR_RED
		}
		COLOR_RESET = '\033[0m'

		return f"{color_mapping.get(level, '')}[{level}] {current_time} :: {message}{COLOR_RESET}"

	def emit_socketio_message(self, message, verbose='INFO'):
		"""
		Emits a message to a Socket.IO instance with a specified verbosity level.

		Args:
			message (str): The message to emit.
			verbose (str): The verbosity level for the message. Defaults to 'INFO'.
		"""
		if self.socketio is not None:
			self.socketio.emit('docker_status', {'message': message, 'verbose': verbose})
		else:
			# Log an error or warning that socketio is not initialized
			print("Attempted to emit a socketio message before socketio was initialized")

	def log(self, message, ws:bool, level=logging.INFO):
		"""
		Logs a message with a specified log level and emits it to a Socket.IO instance.

		Args:
			message (str): The message to log and emit.
			level (int): The log level. Defaults to logging.INFO.
			ws (bool): Emit the message over WebSockets.
		"""
		level_str = logging.getLevelName(level)
		colored_message = self._format_log(level_str, message)
		self.logger.log(level, colored_message)
		if ws:
			self.emit_socketio_message(message, verbose=level_str)

	def trace(self, message, ws=True):
		"""
		Logs a message with the TRACE log level and emits it to a Socket.IO instance.

		Args:
			message (str): The message to log and emit.
			ws (bool): Emit the message over WebSockets. Defaults to True.
		"""
		self.log(message, level=self.TRACE, ws=ws)

	def debug(self, message, ws=True):
		"""
		Logs a message with the DEBUG log level and emits it to a Socket.IO instance.

		Args:
			message (str): The message to log and emit.
			ws (bool): Emit the message over WebSockets. Defaults to True.
		"""
		self.log(message, level=self.DEBUG, ws=ws)

	def info(self, message, ws=True):
		"""
		Logs a message with the INFO log level and emits it to a Socket.IO instance.

		Args:
			message (str): The message to log and emit.
			ws (bool): Emit the message over WebSockets. Defaults to True.
		"""
		self.log(message, level=self.INFO, ws=ws)

	def warn(self, message, ws=True):
		"""
		Logs a message with the WARN log level and emits it to a Socket.IO instance.

		Args:
			message (str): The message to log and emit.
			ws (bool): Emit the message over WebSockets. Defaults to True.
		"""
		self.log(message, level=self.WARN, ws=ws)

	def error(self, message, ws=True):
		"""
		Logs a message with the ERROR log level and emits it to a Socket.IO instance.

		Args:
			message (str): The message to log and emit.
			ws (bool): Emit the message over WebSockets. Defaults to True.
		"""
		self.log(message, level=self.ERROR, ws=ws)


from flask import Blueprint, request, jsonify, Response, url_for
from functools import wraps
from ..services.docker_manager import DockerManager
from ..services.logger import DolostLogger
import os, json

from ..context import Context

# Construct the path to dockerfiles_templates
dockerfiles_path = os.path.join(Context.base_dir, 'dockerfiles_templates')

api_blueprint = Blueprint('api', __name__)

docker_manager = DockerManager.get_instance()
logger = DolostLogger.get_instance()

# Decorator for error handling
def error_handling(func):
    """
    Decorator for error handling.

    Args:
        func (function): The function to be decorated.

    Returns:
        function: The wrapper function for error handling.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            logger.error(f'Error during operation: {str(e)}')
            raise
            return jsonify({"status": "error", "message": str(e)}), 500
    wrapper.__name__ = func.__name__
    return wrapper

# Load existing operations from the JSON file
def load_operations_db():
    """
    Load existing operations from the JSON file.

    Returns:
        list: A list containing operations data loaded from the JSON file.
    """
    try:
        with open('operationsdb.json', 'r') as f:
            operations = json.load(f)
    except FileNotFoundError:
        operations = []
    return operations

# Update the JSON file with modified operations data
def update_operations_db(data):
    """
    Update the JSON file with modified operations data.

    Args:
        data (list): The modified operations data to be written to the JSON file.
    """
    with open('operationsdb.json', 'w') as f:
        json.dump(data, f, indent=4)

# Update an operation based on it's ID
def update_operation_entry(operation_id, new_operation_content, update_decoys:bool=False):
    """
    Update an operation based on its ID.

    Args:
        operation_id (int): The ID of the operation to be updated.
        new_operation_content (dict): The new content for the operation.
        update_decoys (bool, optional): Whether to update decoys data. Defaults to False.
    """
    operations = load_operations_db()
    for operation in operations:
        if operation.get('id') == operation_id:
            # Update the operation content
            for key, value in new_operation_content.items():
                operation[key] = value
            # Update the decoys if update_decoys is True
            if update_decoys:
                operation['decoys'] = new_operation_content.get('decoys', operation.get('decoys', []))
            break
    update_operations_db(operations)

# Modify the operations's decoys 
def update_operation_decoys(operation_id, decoy):
    """
    Modify the operation's decoys.

    Args:
        operation_id (int): The ID of the operation.
        decoy (dict): Decoy information to be added to the operation.
    """
    operations_data = load_operations_db()

    operation = find_operation_by_id(operation_id)

    # Ensure that the "decoys" key holds a list
    if "decoys" not in operation or not isinstance(operation["decoys"], list):
        operation["decoys"] = []
    
    # Filter out any additional fields that are not needed for decoys
    required_decoy_keys = ["Hostname", "Description", "IP", "Subnet", "Gateway", "DeceptionNetwork", "ServicePorts", "Service", "DecoyFiles"]
    filtered_decoys = {key: decoy.get(key) for key in required_decoy_keys}

    # Append the new operation decoys
    operation["decoys"].append(filtered_decoys)
    
    # Write the updated decoys into the file
    return update_operation_entry(operation_id, operation, update_decoys=True)

# Generate ID for new operation based on the last ID in the JSON file
def generate_new_id(operations):
    """
    Generate ID for a new operation based on the last ID in the JSON file.

    Args:
        operations (list): List of existing operations.

    Returns:
        int: The ID for the new operation.
    """
    if not operations:
        return 1
    else:
        # We may need to implement a fail-safe approach
        return operations[-1]['id'] + 1

# Get operation from DB using id
def find_operation_by_id(operation_id):
    """
    Get an operation from the database using its ID.

    Args:
        operation_id (int): The ID of the operation to find.

    Returns:
        dict: The operation data if found, otherwise None.
    """
    operations = load_operations_db()
    for operation in operations:
        if operation.get('id') == operation_id:
            return operation
    return None  # Return None if no operation with the given id is found

# Create a new operation
@api_blueprint.route('/operations/new', methods=['POST'])
@error_handling
def new_operation():
    """
    Create a new operation.

    This endpoint allows the creation of a new operation. It expects a POST request with JSON data containing the necessary information for the new operation.

    Returns:
        jsonify: JSON response indicating the status of the operation creation.
            - status (str): Status of the operation creation ("OK" or "error").
            - message (str): Additional message providing details about the status.
            - redirect (str, optional): URL to redirect the client to upon successful operation creation.

    Raises:
        BadRequest: If the request data is invalid or missing required fields.
    """
    try:
        required_keys = {
            'preparation': ['name', 'objective', 'assets'],
            'narrative': ['storytelling', 'deception_activities', 'monitoring'],
            'closure_criteria': ['limits', 'end_date', 'commander'],
        }

        data = request.get_json()

        # Filter out any additional fields that are not in the required keys
        filtered_data = {}
        for section, keys in required_keys.items():
            if section in data:
                filtered_data[section] = {key: data[section][key] for key in keys if key in data[section]}
            else:
                filtered_data[section] = {}

        # Check if all required keys are present in the filtered data for each section
        missing_keys = []
        for section, keys in required_keys.items():
            missing_keys.extend([f"Missing key '{key}' in section '{section}'" for key in keys if key not in filtered_data[section]])

        if missing_keys:
            return jsonify({"status": "error", "message": "\n".join(missing_keys)}), 400

        # Check if the 'preparation' section is present and 'name' field is not empty
        if not filtered_data['preparation'].get('name'):
            return jsonify({"status": "error", "message": "The 'Operation Name' field under 'Preparation' section cannot be empty"}), 400

        # Check if decoys are provided and filter out any additional keys
        if 'decoys' in data:
            decoys = data['decoys']
            filtered_decoys = []
            for decoy in decoys:
                # Check if all required decoy keys are present
                required_decoy_keys = ["Hostname", "Description", "IP", "Subnet", "Gateway", "DeceptionNetwork", "ServicePorts", "Service", "DecoyFiles"]
                missing_decoy_keys = [key for key in required_decoy_keys if key not in decoy]
                if missing_decoy_keys:
                    return jsonify({"status": "error", "message": f"Missing key(s) {', '.join(missing_decoy_keys)} in decoy"}), 400
                filtered_decoys.append({key: decoy[key] for key in required_decoy_keys})
            filtered_data['decoys'] = filtered_decoys

        new_operation = filtered_data
        actual_operations = load_operations_db()
        
        # Generate ID for the new operation
        new_id = generate_new_id(actual_operations)
        
        # Add the new operation data to the list with the generated ID
        new_operation['id'] = new_id
        actual_operations.append(new_operation)
        update_operations_db(actual_operations)
        
        return jsonify({"status": "OK", "redirect": url_for('views.view_operation', operation_id=new_id)})
    
    except Exception as e:
        return jsonify({"status": "error", "message": str(e)}), 400

# Edit an operation
@api_blueprint.route('/operations/<int:operation_id>/edit', methods=['POST'])
@error_handling
def edit_operation(operation_id):
    """
    Edit an existing operation.

    Args:
        operation_id (int): The ID of the operation to edit.

    Returns:
        jsonify: JSON response indicating the status of the operation editing.
    """
    try:
        required_keys = {
            'preparation': ['name', 'objective', 'assets'],
            'narrative': ['storytelling', 'deception_activities', 'monitoring'],
            'closure_criteria': ['limits', 'end_date', 'commander'],
        }

        data = request.get_json()

        # Filter out any additional fields that are not in the required keys
        filtered_data = {}
        for section, keys in required_keys.items():
            if section in data:
                filtered_data[section] = {key: data[section][key] for key in keys if key in data[section]}
            else:
                filtered_data[section] = {}

        # Check if all required keys are present in the filtered data for each section
        missing_keys = []
        for section, keys in required_keys.items():
            missing_keys.extend([f"Missing key '{key}' in section '{section}'" for key in keys if key not in filtered_data[section]])

        if missing_keys:
            return jsonify({"status": "error", "message": "\n".join(missing_keys)}), 400

        # Check if the 'preparation' section is present and 'name' field is not empty
        if not filtered_data['preparation'].get('name'):
            return jsonify({"status": "error", "message": "The 'Operation Name' field under 'Preparation' section cannot be empty"}), 400

        edited_operation_data = filtered_data

        update_operation_entry(operation_id, edited_operation_data, update_decoys=False)
        
        return jsonify({"status": "OK", "redirect": url_for('views.view_operation', operation_id=operation_id)})
    
    except Exception as e:
        return jsonify({"status": "error", "message": str(e)}), 500

# Remove operation
@api_blueprint.route('/operations/remove/<string:operation_id>', methods=['DELETE'])
@error_handling
def remove_operation(operation_id):
    """
    Remove an operation.

    Args:
        operation_id (str): The ID of the operation to remove.

    Returns:
        jsonify: JSON response indicating the status of the operation removal.
    """
    try:
        operations_data = load_operations_db()

        # Find the operation by ID
        operation_to_remove = find_operation_by_id(int(operation_id))
        if operation_to_remove:
            # Delete the operation data
            operations_data.remove(operation_to_remove)

            update_operations_db(operations_data)

            return jsonify({'message': f'We removed the operation {operation_id}'}), 202
        else:
            return jsonify({'message': f'Operation {operation_id} not found'}), 404

    except Exception:
        return jsonify({'message': f'We can not remove operation {operation_id}'}), 406

# Create a new decoy
@api_blueprint.route('/operations/<int:operation_id>/decoy/new', methods=['POST'])
@error_handling
def new_decoy(operation_id):
    """
    Create a new decoy for an operation.

    Args:
        operation_id (int): The ID of the operation to add the decoy to.

    Returns:
        jsonify: JSON response indicating the status of the new decoy creation.
    """
    data = request.get_json()
    update_operation_decoys(operation_id, data)

    return jsonify({"status": "OK", "message": "New decoy saved"}), 200
      
# Helper function for deployment logic
def deploy_service(image_name, dockerfile_path, decoy_files, service_info):
    """
    Deploy a service using Docker.

    Args:
        image_name (str): The name of the Docker image.
        dockerfile_path (str): The path to the Dockerfile.
        decoy_files (str): The path to service files.
        service_info (dict): Information about the service.

    Returns:
        None
    """
    image_name = docker_manager.build_image(image_name=image_name, decoy_files=decoy_files, dockerfile_path=dockerfile_path)
    docker_manager.run_container(image_name=image_name,**service_info)
    logger.info(f"Deployed container: {service_info['hostname']}")

# Fetch current available Dockerfiles
def get_supported_services():
    """
    Fetches the current available Dockerfile services.

    Returns:
        list: A list of supported services.
    """
    supported_services = []

    for filename in os.listdir(dockerfiles_path):
        if filename.startswith("Dockerfile-"):
            service_name = filename.replace("Dockerfile-", "")
            supported_services.append(service_name)

    return supported_services

@api_blueprint.route('/deploy', methods=['POST'])
@error_handling
def deploy_env():
    """
    Endpoint for deploying the environment.

    Returns:
        Response: A response indicating the deployment status.
    """
    data = request.get_json()
    deploy_collector()
    for decoy in data:
        deploy_decoy(decoy)
    
    return Response("Env Deployed", mimetype='text/plain')

def deploy_collector():
    """
    Deploy the collector service.

    Returns:
        None
    """
    service_dockerfile_path = os.path.join(dockerfiles_path, 'Dockerfile-collector')
    if not os.path.exists(service_dockerfile_path):
        logger.error('Dockerfile not found for Collector')
        return
    
    collector_info = {
        "hostname": "Collector",
        "name": "Collector",
        "network_name": "CollectorNetwork",
        "ipv4_address": "200.100.0.247",
        "subnet": "200.100.0.0/24",
        "gateway": "200.100.0.1",
        "ports": "{'514/tcp': 524, '514/udp': 524 }"
    }
    
    deploy_service(
        image_name="dolost-collector",
        dockerfile_path=service_dockerfile_path,
        decoy_files="collector",
        service_info=collector_info
    )

def deploy_decoy(decoy):
    """
    Deploy a decoy service.

    Args:
        decoy (dict): Decoy information including hostname, network settings, and service details.

    Returns:
        None
    """
    deploy_collector()
    supported_services = get_supported_services()
    decoyservice = decoy["Service"].lower()
    
    if decoyservice not in supported_services:
        logger.error(f"We do not yet offer the service: {decoyservice}")
        return

    decoy_info = {
        "hostname": decoy["Hostname"],
        "name": decoy["Hostname"],
        "network_name": decoy["DeceptionNetwork"],
        "ipv4_address": decoy["IP"],
        "subnet": decoy["Subnet"],
        "gateway": decoy["Gateway"],
        "ports": decoy["ServicePorts"]
    }
    dockerfile_path = os.path.join(dockerfiles_path, f'Dockerfile-{decoyservice}')

    deploy_service(
        image_name=decoyservice,
        dockerfile_path=dockerfile_path,
        decoy_files=decoy["DecoyFiles"],
        service_info=decoy_info
    )

# Remove deployed decoys
@api_blueprint.route('/decoys/clean', methods=['POST'])
@error_handling
def clean_env():
    """
    Endpoint to remove deployed decoys.

    Returns:
        Response: Response indicating the success of the operation.
    """
    # PENDING search for dolost tag
    data = request.get_json()
    for decoy in data:
        clean_decoy(decoy)
    logger.info(f"Removed decoys")
    return Response("Decoys cleaned", mimetype='text/plain')

# Remove decoy
def clean_decoy(decoy):
    """
    Remove decoy.

    Args:
        decoy (dict): Decoy information.

    Returns:
        Response: Response indicating the success of the operation.
    """

    decoy_to_clean = decoy["Hostname"]
    docker_manager.clean_container(decoy_to_clean)

    return Response("A decoy was cleaned", mimetype='text/plain')

# Remove collector
@api_blueprint.route('/collector/clean', methods=['POST'])
@error_handling
def clean_collector():
    """
    Remove collector.

    Returns:
        Response: Response indicating the success of the operation.
    """
    collector_to_clean = "Collector"
    docker_manager.clean_container(collector_to_clean)
    logger.info(f"Removed Collector")
    return Response("Collector cleaned", mimetype='text/plain')

# Remove networks
@api_blueprint.route('/networks/clean', methods=['POST'])
@error_handling
def clean_networks():
    """
    Remove networks.

    Returns:
        Response: Response indicating the success of the operation.
    """
    docker_manager.clean_networks()
    logger.info(f"Removed Networks")
    return Response("Networks cleaned", mimetype='text/plain')

# Start an specific decoy
@api_blueprint.route('/decoys/start', methods=['POST'])
@error_handling
def decoys_start():
    """
    Start a specific decoy.

    Returns:
        Response: JSON response indicating the status of the operation.
    """
    try:
        data = request.get_json()

        docker_manager.start(container_id=data['containerId'])

        return jsonify({"status": "OK", "message": "The container has been started successfully"}), 200

    except Exception as e:
        return jsonify({"status": "error", "message": str(e)}), 500

# Stop an specific decoy
@api_blueprint.route('/decoys/stop', methods=['POST'])
@error_handling
def decoys_stop():
    """
    Stop a specific decoy.

    Returns:
        Response: JSON response indicating the status of the operation.
    """

    try:
        data = request.get_json()

        docker_manager.stop(container_id=data['containerId'])

        return jsonify({"status": "OK", "message": "The container has been stopped successfully"}), 200

    except Exception as e:
        return jsonify({"status": "error", "message": str(e)}), 500

# Deploy an specific decoy, including building and deployment
@api_blueprint.route('/operations/<int:operation_id>/decoys/deploy', methods=['POST'])
@error_handling
def deploy_container(operation_id):
    """
    Deploy a specific decoy, including building and deployment.

    Args:
        operation_id (int): The ID of the operation associated with the decoy.

    Returns:
        Response: JSON response indicating the status of the operation.
    """
    try:
        data = request.get_json()

        operation_data = find_operation_by_id(operation_id)

        decoy_data = None

        for decoy in operation_data.get('decoys', []):
            if "DolosT-" + decoy.get('Hostname') ==  data.get('containerHostname'):
                decoy_data = decoy
                break

        if decoy_data:
            deploy_decoy(decoy_data)
            return jsonify({"status": "OK", "message": "The container has been deployed successfully"}), 200
        else:
            return jsonify({"status": "error", "message": "Decoy not found on stored decoys"}), 404

    except Exception as e:
        logger.error(f"Error deploying container: {str(e)}")
        return jsonify({"status": "error", "message": str(e)}), 500

# Undeploy an specific container decoy
@api_blueprint.route('/operations/<int:operation_id>/decoys/undeploy', methods=['DELETE'])
@error_handling
def undeploy_container(operation_id):
    """
    Undeploy a container decoy for an operation.

    Args:
        operation_id (int): The ID of the operation containing the decoy.

    Returns:
        jsonify: JSON response indicating the status of the container undeployment.
    """
    try:
        data = request.get_json()

        operation_data = find_operation_by_id(operation_id)

        decoy_data = None

        for decoy in operation_data.get('decoys', []):
            if "DolosT-" + decoy.get('Hostname') ==  data.get('containerHostname'):
                decoy_data = decoy
                break

        if decoy_data:
            clean_decoy(decoy_data)
            return jsonify({"status": "OK", "message": "The container has been undeployed successfully"}), 200
        else:
            return jsonify({"status": "error", "message": "Decoy not found on stored decoys"}), 404

    except Exception as e:
        logger.error(f"Error undeploying container: {str(e)}")
        return jsonify({"status": "error", "message": str(e)}), 500

# Delete an specific decoy
@api_blueprint.route('/operations/<int:operation_id>/decoys/delete', methods=['DELETE'])
@error_handling
def delete_decoy(operation_id):
    """
    Delete a specific decoy from an operation.

    Args:
        operation_id (int): The ID of the operation containing the decoy.

    Returns:
        jsonify: JSON response indicating the status of the decoy deletion.
    """
    try:
        data = request.get_json()
        decoy_to_delete = data["decoyHostname"]
        decoy_to_delete = decoy_to_delete.replace("DolosT-", "")
        operation = find_operation_by_id(operation_id)

        operation['decoys'] = [decoy for decoy in operation['decoys'] if decoy["Hostname"] != decoy_to_delete]

        print(operation)

        update_operation_entry(operation_id, operation, update_decoys=True)

        
        return jsonify({"status": "OK", "message": "The decoy has been deleted successfully"}), 200

    except Exception as e:
        logger.error(f"Error deleting decoy: {str(e)}")
        return jsonify({"status": "error", "message": str(e)}), 500

@api_blueprint.route('/config/docker_client', methods=['POST'])
@error_handling
def modify_docker_client():
    """
    Modify the Docker client configuration.

    Returns:
        jsonify: JSON response indicating the status of the Docker client configuration modification.
    """
    def save_file(file_path, content):
        with open(file_path, 'w') as f:
            f.write(content)

    # Check if the folder exists, if not, create it
    if not os.path.exists(Context.CONFIG_FOLDER):
        os.makedirs(Context.CONFIG_FOLDER)

    data = request.get_json()
    
    # Extract data from JSON payload
    config_type = data.get('type')

    dc = {}

    if config_type == "env":
        dc = {'from_env': True}

    elif config_type == "tcp":
        host = data.get('host')
        port = data.get('port')
        dc =  {'tcp': f"tcp://{host}:{port}"}

    elif config_type == "socket":
        socketPath = data.get('socketPath')
        dc = {'socket': socketPath}

    elif config_type == 'tcp_ssl':
        host = data.get('host')
        port = data.get('port')
        ssl_cert = data.get('sslCert')
        ssl_key = data.get('sslKey')
        ssl_ca = data.get('sslCa')
        save_file(os.path.join(Context.CONFIG_FOLDER, 'ssl_cert.pem'), ssl_cert)
        save_file(os.path.join(Context.CONFIG_FOLDER, 'ssl_key.pem'), ssl_key)
        save_file(os.path.join(Context.CONFIG_FOLDER, 'ssl_ca.pem'), ssl_ca)
        dc['tcp_ssl'] = {
            'host': host,
            'port': port,
            'cert_path': os.path.join(Context.CONFIG_FOLDER, 'ssl_cert.pem'),
            'key_path': os.path.join(Context.CONFIG_FOLDER, 'ssl_key.pem'),
            'ca_path': os.path.join(Context.CONFIG_FOLDER, 'ssl_ca.pem')
        }
    else:
        return jsonify({"status": "error", "message": "Invalid configuration sent"}), 400

    try:
        status = docker_manager.check_client_configuration(docker_client=dc)
        if status:        
            docker_manager.configure_client(docker_client=dc)
            docker_manager.check_connection()
            return jsonify({"status": "OK", "message": "Docker client updated successfully"}), 200
        else:
            return jsonify({"status": "error", "message": "Could not connect to docker client"}), 400
    except Exception as e:
        return jsonify({"status": "error", "message": str(e)}), 500
    



from setuptools import setup, find_packages

# Function to read the version from a file
def get_version(rel_path):
    with open(rel_path, "r") as file:
        for line in file.readlines():
            if line.startswith('__version__'):
                # Grabs the first ' or " enclosed string
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")

# Automatically load requirements from a requirements.txt file
def load_requirements(filename='requirements.txt'):
    with open(filename, 'r') as file:
        return [line.strip() for line in file.readlines() if line.strip()]

# Reading the long description from README.md
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='DOLOST',
    version=get_version("src/DOLOST/version.py"),
    author_email="idi@base4sec.com",
    description="Deceptive Operations: Lure, Observe, and Secure Tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Base4Security/DOLOS-T/",
    project_urls={
        "Bug Tracker": "https://github.com/Base4Security/DOLOS-T/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    install_requires=load_requirements(),
    entry_points={
        'console_scripts': [
            'DOLOST = DOLOST.cli:main'
        ]
    },
)
